'''
Created on Nov 20, 2017

@author: Charlie McLouth
'''

from flask.config import Config
from flask import Flask
import flask.json
import os.path
from p4rest.util import osenviron, osmakedirs, osstat
import logging

logger = logging.getLogger(__name__)

class P4RestAppConfig(Config):
    '''Object for loading, dumping, and validating configuration information'''

    managed_keys = ["WORKDIR", "P4TRUST", "P4CLIENT", "P4TICKETS", "P4CONFIG", 
                    "MAXRESULTS", "SESSION_TIMEOUT"]

    def get_config_file(self, app_config_key):
        configfile = osenviron.get(app_config_key, self.get(app_config_key, None))
        if configfile is None:
            raise RuntimeError("The environment variable {} is not set and as "
                               "such configuration could not be loaded.  Set "
                               "this variable and make it point to a json "
                               "encoded configuration file"
                               "".format(app_config_key))
        return configfile

    def set_program_defaults(self, app_path=None):
        '''set sensible default values for required configuration settings'''
        logger.debug("set_program_defaults() enter")
        if app_path is None:
            app_path = self.root_path
        k = "WORKDIR"
        defaultvalue = os.path.abspath(os.path.join(app_path, "workdir"))
        self.setdefault(k, defaultvalue)
        k = "P4TRUST"
        defaultvalue = ".p4trust"
        self.setdefault(k, defaultvalue)
        k = "P4TICKETS"
        defaultvalue = ".p4tickets"
        self.setdefault(k, defaultvalue)
        k = "P4CONFIG"
        defaultvalue = ".session"
        self.setdefault(k, defaultvalue)
        k = "P4CLIENT"
        defaultvalue = "_____CLIENT_UNSET_____"
        self.setdefault(k, defaultvalue)
        k = "MAXRESULTS"
        defaultvalue = 1000
        self.setdefault(k, defaultvalue)
        k = "SESSION_TIMEOUT"
        defaultvalue = 300
        self.setdefault(k, defaultvalue)
        # This is required by flask_restful so that it uses the correct encoder
        k = "RESTFUL_JSON"
        if k not in self:
            self[k] = {"cls": Flask.json_encoder}
        else:
            self[k]["cls"] = Flask.json_encoder
        logger.debug("set_program_defaults() exit")

    def dump_to_json(self, filename, silent=True):
        '''dump the application specific values to json file'''
        logger.debug("dump_to_json() enter")
        dictToWrite = {}
        workdir = self.get("WORKDIR", None)
        for k in P4RestAppConfig.managed_keys:
            v = self.get(k, None)
            if v is not None:
                # convert physical paths to relative paths
                if workdir is not None and k in ["P4TRUST", "P4TICKETS"] \
                        and os.path.isabs(v) and v.startswith(workdir):
                    test = os.path.relpath(v, workdir)
                    if not test.startswith(".."):
                        v = test
                dictToWrite[k] = v
        try:
            with open(filename, 'w') as f:
                flask.json.dump(dictToWrite, f, sort_keys=True, indent=4)
            last_mod_file = osstat(filename).st_mtime
            if last_mod_file is not None:
                self["CONFIGFILE_LASTMOD"] = last_mod_file
        except Exception as e:
            if silent:
                logger.debug("suppressed exception: " + e.strerror, 
                             exc_info=True)
            else:
                raise e
        logger.debug("dump_to_json() exit")

    def validate(self, app_config_key):
        '''validate the application configuration'''
        logger.debug("validate() enter")
        self.get_config_file(app_config_key)
        for k in P4RestAppConfig.managed_keys:
            v = self.get(k, None)
            if v is None:
                raise RuntimeError("Configuration item not defined: {}"
                                   "".format(k))
            if k == "WORKDIR":
                if not os.path.isabs(v):
                    raise RuntimeError("Invalid value for configuration item "
                                       "value '{}': {}".format(k, v))
                if not os.path.isdir(v):
                    osmakedirs(v, exist_ok=True)
            elif k in ["P4TRUST", "P4TICKETS", "P4CONFIG"]:
                if k != "P4CONFIG":
                    if not os.path.isabs(v):
                        v = os.path.join(self["WORKDIR"], v)
                    dirname = os.path.dirname(v)
                    if not os.path.isdir(dirname):
                        osmakedirs(dirname, exist_ok=True)
                    if not os.path.isfile(v):
                        with open(v, "a"):
                            pass
                    # convert to physical path
                    if v != self[k]:
                        self[k] = v
                # set environment variables
                oldvalue = osenviron.get(k, None)
                if oldvalue is None or oldvalue != v:
                    osenviron[k] = v
        logger.debug("validate() exit")

    def from_json_if(self, app_config_key, silent=False):
        '''load configuration from json if and only if the file has changed'''
        logger.debug("from_json_if() enter")
        bLoaded = False
        configfile = self.get_config_file(app_config_key)
        if os.path.exists(configfile):
            last_mod = self.get("CONFIGFILE_LASTMOD", None)
            last_mod_file = osstat(configfile).st_mtime
            bLoaded = last_mod is None
            if not bLoaded:
                bLoaded = last_mod_file is None or last_mod_file > last_mod
        else:
            bLoaded = silent
        if bLoaded:
            Config.from_json(self, configfile, silent)
            self["CONFIGFILE_LASTMOD"] = osstat(configfile).st_mtime
        logger.debug("from_json_if() exit")
        return bLoaded
